// -*- C++ -*-
/*!
 * @file  RT_Gesture_HouseControl.cpp
 * @brief Gesture_HouseControl
 * @date $Date$
 *
 * $Id$
 */

#include "RT_Gesture_HouseControl.h"
#include "stdio.h"
#include "math.h"

#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>
#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/idl/ExtendedDataTypesSkel.h>

int timer_counter = 0;//vOs100[ms]
TimedOctetSeq old_m_module_control;
int old_flag_position;
// Module specification
// <rtc-template block="module_spec">
static const char* rt_gesture_housecontrol_spec[] =
  {
    "implementation_id", "RT_Gesture_HouseControl",
    "type_name",         "RT_Gesture_HouseControl",
    "description",       "Gesture_HouseControl",
    "version",           "1.0.0",
    "vendor",            "AIST",
    "category",          "Category",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    //"exec_cxt.periodic.rate", "1.0",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
RT_Gesture_HouseControl::RT_Gesture_HouseControl(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_target_headIn("target_head", m_target_head),
    m_target_neckIn("target_neck", m_target_neck),
    m_target_left_shoulderIn("target_left_shoulder", m_target_left_shoulder),
    m_target_left_elbowIn("target_left_elbow", m_target_left_elbow),
    m_target_left_handIn("target_left_hand", m_target_left_hand),
    m_target_right_shoulderIn("target_right_shoulder", m_target_right_shoulder),
    m_target_right_elbowIn("target_right_elbow", m_target_right_elbow),
    m_target_right_handIn("target_right_hand", m_target_right_hand),
    m_string_controlIn("string_control", m_string_control),
    m_data_in_controlIn("data_in_control", m_data_in_control),
    m_house_controlOut("House_Control", m_house_control),
    m_module_controlOut("Module_control", m_module_control),
    m_voice_controlOut("voice_control", m_voice_control),
    m_add_controlOut("add_control", m_add_control)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
RT_Gesture_HouseControl::~RT_Gesture_HouseControl()
{
}



RTC::ReturnCode_t RT_Gesture_HouseControl::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  registerInPort("target_head", m_target_headIn);
  registerInPort("target_neck", m_target_neckIn);
  registerInPort("target_left_shoulder", m_target_left_shoulderIn);
  registerInPort("target_left_elbow", m_target_left_elbowIn);
  registerInPort("target_left_hand", m_target_left_handIn);
  registerInPort("target_right_shoulder", m_target_right_shoulderIn);
  registerInPort("target_right_elbow", m_target_right_elbowIn);
  registerInPort("target_right_hand", m_target_right_handIn);
  registerInPort("string_control", m_string_controlIn);
  registerInPort("data_in_control", m_data_in_controlIn);
  
  // Set OutPort buffer
  registerOutPort("House_Control", m_house_controlOut);
  registerOutPort("Module_control", m_module_controlOut);
  registerOutPort("voice_control", m_voice_controlOut);
  registerOutPort("add_control", m_add_controlOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  return RTC::RTC_OK;
}

double theata_cal(TimedPoint3D point0, TimedPoint3D point1)
{
	double theata;

	if((point1.data.x >= point0.data.x)&&(point1.data.z >= point0.data.z)) 
	{
		theata = atan((point1.data.x - point0.data.x) / (point1.data.z - point0.data.z));
	}
	else if((point1.data.x >= point0.data.x)&&(point0.data.z >= point1.data.z))
	{
		theata = atan((point1.data.z - point0.data.z) / (point0.data.x - point1.data.x));
		theata += 3.141592 * 0.5;
	}
	else if((point0.data.x >= point1.data.x)&&(point0.data.z >= point1.data.z))
	{
		theata = atan((point0.data.x - point1.data.x) / (point0.data.z - point1.data.z));
		theata += 3.141592;
	}
	else if((point0.data.x >= point1.data.x)&&(point1.data.z >= point0.data.z))
	{
		theata = atan((point1.data.z - point0.data.z) / (point0.data.x - point1.data.x));
		theata += 3.141592*1.5;
	}

	return(theata) ;
}

double eta_cal(TimedPoint3D point0, TimedPoint3D point1)
{
	double eta;

	if((point1.data.x >= point0.data.x)&&(point0.data.y >= point1.data.y)) 
	{
		eta = atan((point1.data.x - point0.data.x) / (point0.data.y - point1.data.y));
	}
	else if((point1.data.x >= point0.data.x)&&(point1.data.y >= point0.data.y))
	{
		eta = atan((point1.data.y - point0.data.y) / (point1.data.x - point0.data.x));
		eta += 3.141592*0.5;
	}
	else if((point0.data.x >= point1.data.x)&&(point1.data.y >= point0.data.y))
	{
		eta = atan((point0.data.x - point1.data.x) / (point1.data.y - point0.data.y)); 
		eta += 3.141592;
	}
	else if((point0.data.x >= point1.data.x)&&(point0.data.y >= point1.data.y))
	{
		eta = atan((point0.data.y - point1.data.y) / (point0.data.x - point1.data.x)); 
		eta += 3.141592*1.5;
	}

	return(eta) ;
}

TimedPoint3D Room_Position_Cal(TimedPoint3D point)
{
	TimedPoint3D room_position;
	double kinect_position_theata_rad = 3.141592 * KINECT_POSITION_THETA/180;
	double alpha;
	double r_length;
/*#define KINECT_POSITION_X 3300 //̒[_ƂƂKinecẗʒux[mm]
#define KINECT_POSITION_Y 700 //̒[_ƂƂKinecẗʒuy[mm]
#define KINECT_POSITION_Z 1300 //̒[_ƂƂKinecẗʒuz[mm]
#define KINECT_POSITION_THETA 115 //Kinectݒupx@ɑ΂ĐAvɐ[deg]
*/
	alpha = atan(point.data.x/point.data.z);
	alpha += kinect_position_theata_rad;
	r_length = sqrt((point.data.x * point.data.x)+(point.data.z * point.data.z));
	room_position.data.x = KINECT_POSITION_X + (r_length * cos(alpha));
	room_position.data.z = KINECT_POSITION_Z + (r_length * sin(alpha));
	room_position.data.y = point.data.y + KINECT_POSITION_Y;
	return(room_position);
}

TimedPoint3D Hand_Endpoint_Cal(TimedPoint3D point1, TimedPoint3D point2)
{
	TimedPoint3D hand_end_point;
	double a,b;
	double x,z;
	double delta_x, delta_z;
	double temp_x, temp_z;
	double temp2_x, temp2_z;
	double tau;

	hand_end_point.data.y = point2.data.y;

	delta_x = abs(point1.data.x - point2.data.x);
	delta_z = abs(point1.data.z - point2.data.z);
/*	if(delta_x < 50)
	{
		if(point1.data.z > point2.data.z)
		{
			hand_end_point.data.x = point2.data.x;
			hand_end_point.data.z = 0;
		}
		else if(point2.data.z > point1.data.z)
		{
			hand_end_point.data.x = point1.data.x;
			hand_end_point.data.z = ROOM_Z;
		}
	}
	else if(delta_z < 50)
	{
		if(point1.data.x > point2.data.x)
		{
			hand_end_point.data.z = point2.data.z;
			hand_end_point.data.x = 0;
		}
		else if(point2.data.x > point1.data.x)
		{
			hand_end_point.data.z = point1.data.z;
			hand_end_point.data.x = ROOM_X;
		}
	}
	else
	{
*/
		a = ((point1.data.z - point2.data.z)/(point1.data.x - point2.data.x));
		b = point1.data.z - (a * point1.data.x);

		if((point2.data.x > point1.data.x)&&(point1.data.z > point2.data.z))
		{
			temp_x = -b/a;
			temp_z = (a * ROOM_X) + b;
			tau = atan((point2.data.x - point1.data.x)/(point1.data.z - point2.data.z));
			temp2_x = (point1.data.z * tan(tau)) + point1.data.x;
			if((0 < temp_x)&&(temp_x < ROOM_X))
			{
				hand_end_point.data.x = temp_x;
				hand_end_point.data.z = 0;
			}
			else if((tau < (PAI/4))&&(temp2_x < ROOM_X)) 
			{
				hand_end_point.data.x = temp2_x;
				hand_end_point.data.z = 0;
			}
			else 
			{
				hand_end_point.data.x = ROOM_X;
				hand_end_point.data.z = temp_z;
			}
		}
		else if((point2.data.x > point1.data.x)&&(point2.data.z > point1.data.z))
		{
			temp_x = (ROOM_Z - b)/a;
			temp_z = (a * ROOM_X) + b;
			tau = atan((point2.data.z - point1.data.z)/(point2.data.x - point1.data.x));
			temp2_x = point1.data.x + (ROOM_Z - point1.data.z) * tan(tau);
			if((0 < temp_z)&&(temp_z < ROOM_Z))
			{
				hand_end_point.data.x = ROOM_X;
				hand_end_point.data.z = temp_z;
			}
			else if((tau < (PAI/4))&&(temp2_x < ROOM_X)) 
			{
				hand_end_point.data.x = temp2_x;
				hand_end_point.data.z = ROOM_Z;
			}
			else
			{
				hand_end_point.data.x = temp_x;
				hand_end_point.data.z = ROOM_Z;
			}
		}
		else if((point1.data.x > point2.data.x)&&(point2.data.z > point1.data.z))
		{
			temp_x = (ROOM_Z - b)/a;
			temp_z = b;
			tau = atan((point1.data.x - point2.data.x)/(point2.data.z - point1.data.z));
			temp2_x = point1.data.x - (ROOM_Z - point1.data.z) * tan(tau);
			if((0 < temp_z)&&(temp_z < ROOM_Z))
			{
				hand_end_point.data.x = 0;
				hand_end_point.data.z = temp_z;
			}
			else if((tau < (PAI/4))&&(temp2_x > 0)) 
			{
				hand_end_point.data.x = temp2_x;
				hand_end_point.data.z = ROOM_Z;
			}
			else
			{
				hand_end_point.data.x = temp_x;
				hand_end_point.data.z = ROOM_Z;
			}
		}
		else if((point1.data.x > point2.data.x)&&(point1.data.z > point2.data.z))
		{
			temp_x = -b/a;
			temp_z = b;
			tau = atan((point1.data.x - point2.data.x)/(point1.data.z - point2.data.z));
			temp2_x = point1.data.x - point1.data.z * tan(tau);
			if((0 < temp_z)&&(temp_z < ROOM_Z))
			{
				hand_end_point.data.x = 0;
				hand_end_point.data.z = temp_z;
			}
			else if((tau < (PAI/4))&&(temp2_x > 0)) 
			{
				hand_end_point.data.x = temp2_x;
				hand_end_point.data.z = 0;
			}
			else
			{
				hand_end_point.data.x = temp_x;
				hand_end_point.data.z = 0;
			}
		}
	//}

	return(hand_end_point);
}

TimedOctetSeq ObjectSelect(TimedPoint3D arm_end_point)
{
	TimedOctetSeq object;
	object.data.length(24);
	for(int i=0; i<23 ;i++) 	object.data[i] = 0x00;//ftHg
	object.data[22] = 0x00;//IvV1
	object.data[23] = 0x00;//IvV2

	if(((arm_end_point.data.x > DigitalTelevisionID1_x1)&&(arm_end_point.data.x < DigitalTelevisionID1_x2))||((arm_end_point.data.z > DigitalTelevisionID1_z1)&&(arm_end_point.data.z < DigitalTelevisionID1_z2)))
	{
		object.data[12] = AudioVisual_1;//IuWFNg1
		object.data[13] = AudioVisual_2;//IuWFNg2
		object.data[14] = DigitalTelevision_1;//IuWFNg1
		object.data[15] = DigitalTelevision_2;//IuWFNg2
		object.data[16] = DigitalTelevisionID1_1;//foCXԍ1
		object.data[17] = DigitalTelevisionID1_2;//foCXԍ2
		object.data[18] = Power_1;//썀1
		object.data[19] = Power_2;//썀2
		object.data[20] = Toggle_1;//e1
		object.data[21] = Toggle_2;//e2
	}
	return(object);
}

/*
RTC::ReturnCode_t RT_Gesture_HouseControl::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RT_Gesture_HouseControl::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RT_Gesture_HouseControl::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t RT_Gesture_HouseControl::onActivated(RTC::UniqueId ec_id)
{
	m_target_head.data.x = m_target_head.data.y = m_target_head.data.z =0;

	m_target_head = m_target_head;
    m_target_neck = m_target_head;
    m_target_left_shoulder = m_target_head;
    m_target_left_elbow = m_target_head;
    m_target_left_hand = m_target_head;
    m_target_right_shoulder = m_target_head;
    m_target_right_elbow = m_target_head;
    m_target_right_hand = m_target_head;
		
	m_module_control.data.length(24);
	for(int i=0; i<24 ;i++) 	m_module_control.data[i] = 0x00;//ftHg
	old_m_module_control.data.length(24);
	for(int i=0; i<24 ;i++) 	old_m_module_control.data[i] = 0x00;//ftHg

	old_flag_position = 0;
	return RTC::RTC_OK;
}


RTC::ReturnCode_t RT_Gesture_HouseControl::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RT_Gesture_HouseControl::onExecute(RTC::UniqueId ec_id)
{
	double theata_left_shoulder_elbow, theata_left_elbow_hand, theata_left_shoulder_hand;
	double theata_right_shoulder_elbow, theata_right_elbow_hand, theata_right_shoulder_hand;
	double eta_left_shoulder_elbow, eta_left_elbow_hand, eta_left_shoulder_hand;
	double eta_right_shoulder_elbow, eta_right_elbow_hand, eta_right_shoulder_hand;

	double left_hand_theata, left_hand_eta;
	double right_hand_theata, right_hand_eta;
	TimedPoint3D room_left_shoulder, room_left_elbow, room_left_hand;
	TimedPoint3D room_right_shoulder, room_right_elbow, room_right_hand;

	int flag_position = 0;

	TimedPoint3D left_hand_end, right_hand_end;
	TimedPoint3D crap_hand;

	m_target_headIn.read(); 
    m_target_neckIn.read();
    m_target_left_shoulderIn.read();
    m_target_left_elbowIn.read();
    m_target_left_handIn.read();
    m_target_right_shoulderIn.read();
    m_target_right_elbowIn.read();
    m_target_right_handIn.read();
    m_string_controlIn.read();
    m_data_in_controlIn.read();

	theata_left_shoulder_elbow = theata_cal(m_target_left_shoulder, m_target_left_elbow);
	theata_left_elbow_hand = theata_cal(m_target_left_elbow, m_target_left_hand);
	theata_right_shoulder_elbow = theata_cal(m_target_right_shoulder, m_target_right_elbow);
	theata_right_elbow_hand = theata_cal(m_target_right_elbow, m_target_right_hand);
	eta_left_shoulder_elbow = eta_cal(m_target_left_shoulder, m_target_left_elbow);
	eta_left_elbow_hand = eta_cal(m_target_left_elbow, m_target_left_hand);
	eta_right_shoulder_elbow = eta_cal(m_target_right_shoulder, m_target_right_elbow);
	eta_right_elbow_hand = eta_cal(m_target_right_elbow, m_target_right_hand);

	left_hand_theata = abs(theata_left_shoulder_elbow - theata_left_elbow_hand);
	right_hand_theata = abs(theata_right_shoulder_elbow - theata_right_elbow_hand);
	left_hand_eta = abs(eta_left_shoulder_elbow - eta_left_elbow_hand);
	right_hand_eta = abs(eta_right_shoulder_elbow - eta_right_elbow_hand);

	room_left_shoulder = Room_Position_Cal(m_target_left_shoulder);
	room_left_elbow = Room_Position_Cal(m_target_left_elbow);
	room_left_hand = Room_Position_Cal(m_target_left_hand);
	room_right_shoulder = Room_Position_Cal(m_target_right_shoulder);
	room_right_elbow = Room_Position_Cal(m_target_right_elbow);
	room_right_hand = Room_Position_Cal(m_target_right_hand);

	//printf("right_elbow  x*%f, y:%f, z:%f\n", m_target_right_elbow.data.x, m_target_right_elbow.data.y, m_target_right_elbow.data.z); 
	//printf("right_hand  x*%f, y:%f, z:%f\n", m_target_right_hand.data.x, m_target_right_hand.data.y, m_target_right_hand.data.z); 
	//printf("left_elbow  x*%f, y:%f, z:%f\n", m_target_left_elbow.data.x, m_target_left_elbow.data.y, m_target_left_elbow.data.z); 
	//printf("left_hand  x*%f, y:%f, z:%f\n", m_target_left_hand.data.x, m_target_left_hand.data.y, m_target_left_hand.data.z); 
	//printf("right_elbow  x*%f, y:%f, z:%f\n", m_target_right_elbow.data.x, m_target_right_elbow.data.y, m_target_right_elbow.data.z); 
	//printf("right_hand  x*%f, y:%f, z:%f\n", m_target_right_hand.data.x, m_target_right_hand.data.y, m_target_right_hand.data.z); 
/*
	printf("//////\n");
	printf("---\n");
	printf("room_left_elbow  x:%f, y:%f, z:%f\n", room_left_elbow.data.x, room_left_elbow.data.y, room_left_elbow.data.z); 
	printf("room_left_hand  x:%f, y:%f, z:%f\n", room_left_hand.data.x, room_left_hand.data.y, room_left_hand.data.z); 
	printf("room_right_elbow  x:%f, y:%f, z:%f\n", room_right_elbow.data.x, room_right_elbow.data.y, room_right_elbow.data.z); 
	printf("room_right_hand  x:%f, y:%f, z:%f\n", room_right_hand.data.x, room_right_hand.data.y, room_right_hand.data.z); 
*/

	
	left_hand_end = Hand_Endpoint_Cal(room_left_shoulder, room_left_hand);
	right_hand_end = Hand_Endpoint_Cal(room_right_shoulder, room_right_hand);
	printf("left_hand_end  x:%f, y:%f, z:%f\n", left_hand_end.data.x, left_hand_end.data.y, left_hand_end.data.z); 
	printf("right_hand_end  x:%f, y:%f, z:%f\n", right_hand_end.data.x, right_hand_end.data.y, right_hand_end.data.z); 

	//printf("theata_left_shoulder_elbow = %f\n", theata_left_shoulder_elbow);
	//printf("theata_left_elbow_hand = %f\n", theata_left_elbow_hand);
	//printf("theata_right_shoulder_elbow = %f\n", theata_right_shoulder_elbow);
	//printf("theata_right_elbow_hand = %f\n", theata_right_elbow_hand);
	//printf("eta_left_shoulder_elbow = %f\n", eta_left_shoulder_elbow);
	//printf("eta_left_elbow_hand = %f\n", eta_left_elbow_hand);
	//printf("eta_right_shoulder_elbow = %f\n", eta_right_shoulder_elbow);
	//printf("eta_right_elbow_hand = %f\n", eta_right_elbow_hand);

	//printf("right_elbow  x*%f, y:%f, z:%f\n", m_target_right_elbow.data.x, m_target_right_elbow.data.y, m_target_right_elbow.data.z); 
	//printf("right_hand  x*%f, y:%f, z:%f\n", m_target_right_hand.data.x, m_target_right_hand.data.y, m_target_right_hand.data.z); 
/*	printf("theata_left_shoulder_elbow = %f\n", theata_left_shoulder_elbow);
	printf("theata_left_elbow_hand = %f\n", theata_left_elbow_hand);
	printf("theata_right_shoulder_elbow = %f\n", theata_right_shoulder_elbow);
	printf("theata_right_elbow_hand = %f\n", theata_right_elbow_hand);
	printf("eta_left_shoulder_elbow = %f\n", eta_left_shoulder_elbow);
	printf("eta_left_elbow_hand = %f\n", eta_left_elbow_hand);
	printf("eta_right_shoulder_elbow = %f\n", eta_right_shoulder_elbow);
	printf("eta_right_elbow_hand = %f\n", eta_right_elbow_hand);
*/
	if((left_hand_theata < PAI/4) && (left_hand_eta < PAI/4)&&(eta_left_elbow_hand>PAI/4)&&(eta_left_elbow_hand<DEG315)) 
	{
		printf("Left elbow is straighten\n");
		if((eta_left_elbow_hand > PAI/2)&&(eta_left_elbow_hand < DEG270))
		{
			//printf("Left Hand UP\n");
			flag_position +=1;
		}
		m_module_control = ObjectSelect(left_hand_end);
		if(m_module_control.data[15] == old_m_module_control.data[15])
		{
			timer_counter++;
			if(timer_counter > 25) 
			{
				m_module_controlOut.write();
				Sleep(2000);
				timer_counter = 0;
			}
		}
		else
		{
			timer_counter = 0;
		}
		old_m_module_control = m_module_control;
	}
	if((right_hand_theata < PAI/4) && (right_hand_eta < PAI/4)&&(eta_right_elbow_hand>PAI/4)&&(eta_right_elbow_hand<DEG315))
	{
		//printf("Right elbow is straighten\n");
		if((eta_right_elbow_hand > PAI/2)&&(eta_right_elbow_hand < DEG270))
		{
			//printf("Right Hand UP\n");
			flag_position +=2;
		}
		if((eta_right_elbow_hand > DEG200)&&(eta_right_elbow_hand < DEG270))
		{
			//printf("Right Hand UP\n");
			flag_position = 4;
		}
	}
	if(old_flag_position != flag_position)
	{
		m_add_control.data = flag_position;
		m_add_controlOut.write();
	}
	old_flag_position = flag_position;
	//printf("Flag Position = %d\n",m_add_control.data);
	//printf("---\n");
	
	return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RT_Gesture_HouseControl::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t RT_Gesture_HouseControl::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RT_Gesture_HouseControl::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RT_Gesture_HouseControl::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RT_Gesture_HouseControl::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void RT_Gesture_HouseControlInit(RTC::Manager* manager)
  {
    coil::Properties profile(rt_gesture_housecontrol_spec);
    manager->registerFactory(profile,
                             RTC::Create<RT_Gesture_HouseControl>,
                             RTC::Delete<RT_Gesture_HouseControl>);
  }
  
};


